/*
	various bits and pieces gathered from euphoria archives
*/
namespace clib
constant __MODULENAME__ = "clib"

include get.e

include std/dll.e
include std/machine.e

public function lo_byte(integer i1)
    return and_bits(i1, #FF)
end function

public function hi_byte(integer i1)
    return floor(and_bits(i1,#FF00) / #100)
end function

public function lo_word(atom a)
    return and_bits(a, #FFFF)
end function

public function hi_word(atom a)
    return and_bits(floor( a / #10000 ), #FFFF)
end function

public function make_word( integer lo, integer hi)
   lo  = and_bits( lo,  #FF )
   hi = and_bits( hi, #FF )
   return hi * #100 + lo
end function

public function make_long( integer lo, integer hi)
   lo  = and_bits( lo,  #FFFF )
   hi = and_bits( hi, #FFFF )
   return hi * #10000 + lo
end function

public function bitshift(atom x, atom shift) 
	return floor(x * power(2,shift))
end function  

public function leftshift(atom x, atom shift) 
	return bitshift(x,shift) 
end function  

public function rightshift(atom x, atom shift) 
	return bitshift(x,-shift) 
end function  

public function link_dll(sequence name, sequence path="")
	atom     handle
	sequence sodll_ext = ".so"
	ifdef WIN32 then 
		sodll_ext = ".dll"
	end ifdef
	name   = path & name & sodll_ext
	ifdef WIN32 then
		handle = open_dll( name )
	elsedef
		sequence checkpath = { "./", "/usr/lib/", "/usr/local/lib/" }
		for i = 1 to length(checkpath) label "start" do
			handle = open_dll(checkpath[i] & name)
			if handle > 0 then
				exit "start"
			end if
		end for
	end ifdef
	if handle = NULL then
		puts(1, __MODULENAME__ & " : could not load shared library/dll '" & name & "'\n")
		abort(0)
	end if
	return handle
end function

public function link_func(atom dll, sequence name, sequence args, atom result, integer quit=1)
	integer handle
	handle = define_c_func(dll, name, args, result)
	if handle = -1 then
		puts(1, __MODULENAME__ & " : could not link to C function '" & name & "'\n")
		if (quit) then
			abort(0)
		end if
	end if
	return handle
end function

public function link_proc(atom dll, sequence name, sequence args, integer quit=1)
	integer handle
	handle = define_c_proc(dll, name, args)
	if handle = -1 then
		puts(1, __MODULENAME__ & " : could not link to C procedure '" & name & "'\n")
		if (quit) then
			abort(0)
		end if
	end if
	return handle
end function

public function link_var(atom dll, sequence name, integer quit=1)
	atom link = define_c_var(dll, name)
	if (link = -1) then
		puts(1, __MODULENAME__ & " : could not link C variable '" & name & "'\n")
		if (quit) then
			abort(0)
		end if
	end if
	return link
end function

--// RAS Dec 09
--// available/free memory

public enum 
	memLength,
	memMemoryLoad,
	memTotalPhys,
	memAvailPhys,
	memTotalPageFile,
	memAvailPageFile,
	memTotalVirtual,
	memAvailVirtual,
	memAvailExtendedVirtual,
	--// linux etc ...
	memPageSize,
	memNumPages,
	memAvNumPages

constant sizeof_memstatus   = 32
constant sizeof_memstatusex = 64

ifdef WIN32 then
	/*
		sizeof(DWORD)          = 4
		sizeof(DWORDLONG)      = 8
		sizeof(MEMORYSTATUS)   = 32
		sizeof(MEMORYSTATUSEX) = 64

		MEMORYSTATUS 
			DWORD dwLength;
			DWORD dwMemoryLoad;
			DWORD dwTotalPhys;
			DWORD dwAvailPhys;
			DWORD dwTotalPageFile;
			DWORD dwAvailPageFile;
			DWORD dwTotalVirtual;
			DWORD dwAvailVirtual;
		MEMORYSTATUSEX
			DWORD     dwLength;
			DWORD     dwMemoryLoad;
			DWORDLONG ullTotalPhys;
			DWORDLONG ullAvailPhys;
			DWORDLONG ullTotalPageFile;
			DWORDLONG ullAvailPageFile;
			DWORDLONG ullTotalVirtual;
			DWORDLONG ullAvailVirtual;
			DWORDLONG ullAvailExtendedVirtual;
	*/
	constant clib    = open_dll("kernel32")
	constant memconf = define_c_proc(clib, "GlobalMemoryStatus",   { C_POINTER })
	-- constant memconf = define_c_proc(clib, "GlobalMemoryStatusEx", { C_POINTER })
elsifdef LINUX then
	constant clib = open_dll("")
elsifdef FREEBSD or SUNOS or OPENBSD then
	constant clib = open_dll("libc.so")
elsifdef OSX then
	constant clib = open_dll("libc.dylib")
end ifdef

ifdef not WIN32 then
	constant SC_PAGESIZE     = 30
	constant SC_PHYS_PAGES   = 85
	constant SC_AVPHYS_PAGES = 86
	constant memconf         = define_c_func(clib, "sysconf", { C_INT }, C_LONG)
end ifdef

constant PAGEDIV = 1024

public function sysconf(integer id)
	ifdef not WIN32 then
		if not memconf then
			return 0
		end if
		return c_func(memconf, { id })
	end ifdef
	return -1
end function

public function memstatus(integer in_mb=1)
	sequence sysmem = repeat(0, memAvNumPages)
	if not memconf then
		return sysmem
	end if
	ifdef WIN32 then
		sequence m    = sysmem[memLength..memAvailVirtual]
		m[memLength]  = sizeof_memstatus
		atom     stat = allocate(sizeof(int) * length(m))
		c_proc(memconf, { stat })
		sysmem[memLength..memAvailVirtual] = peek4u({ stat, length(m) })
		sysmem[memTotalPhys..$] = sysmem[memTotalPhys..$]
		free(stat)
	elsedef
		sysmem[memPageSize]   = sysconf(SC_PAGESIZE)
		sysmem[memNumPages]   = sysconf(SC_PHYS_PAGES)
		sysmem[memAvNumPages] = sysconf(SC_AVPHYS_PAGES)
		sysmem[memTotalPhys]  = sysmem[memNumPages]   * sysmem[memPageSize]
		sysmem[memAvailPhys]  = sysmem[memAvNumPages] * sysmem[memPageSize]
	end ifdef
	--// want size in MB
	if (in_mb) then
		sysmem[memTotalPhys] /= PAGEDIV
		sysmem[memAvailPhys] /= PAGEDIV
	end if
	return sysmem
end function

public function ipaddr_to_long(sequence ipADDR)
	sequence v, ipQUAD = repeat(0,4)
	integer i = 1, p = match(".", ipADDR)
	while p > 0 do
		v         = value(ipADDR[1..p - 1])
		ipQUAD[i] = v[$]
		i      += 1
		ipADDR  = ipADDR[p + 1..$]
		p       = match(".", ipADDR)
	end while
	v         = value(ipADDR)
	ipQUAD[i] = v[$]
 	return ((((((ipQUAD[1]) * 256) + (ipQUAD[2])) * 256) + (ipQUAD[3])) * 256) + (ipQUAD[4])
end function

public function long_to_ipaddr(atom longIP)
	sequence ipADDR = sprintf("%d", { remainder(longIP, 256) })
	for i = 1 to 3 do
		longIP = floor(longIP / 256)
		ipADDR = sprintf("%d", { remainder(longIP,256) }) & "." & ipADDR
	end for
	return ipADDR
end function

ifdef __TEST_CLIB__ then

	atom longIP = ipaddr_to_long("192.168.11.150")
	printf(1, "%d = %s\n", { longIP, long_to_ipaddr(longIP) })

	longIP = ipaddr_to_long("192.168.0.5")
	printf(1, "%d = %s\n", { longIP, long_to_ipaddr(longIP) })

	longIP = ipaddr_to_long("127.0.0.1")
	printf(1, "%d = %s\n", { longIP, long_to_ipaddr(longIP) })

	longIP = -1062731771)
	printf(1, "%d = %s\n", { longIP, long_to_ipaddr(longIP) })

end ifdef
