/*
	desc      : Mongoose Embedded Web Server
	author    : Raphael Smart
	date      : May 2012

	licence/disclaimer:

	This software is provided 'as-is', without any express or implied warranty.
	In no event will the author be held liable for any damages arising from
	the use of this software.
*/

namespace libmongoose
constant __MODULENAME__ = "libmongoose"

include std/machine.e
include std/dll.e

include lib/clib.e

ifdef WINDOWS then
	constant SOEXT = "dll"
elsedef
	constant SOEXT = "so"
end ifdef

/* imports */

constant libmongoose_lib = link_dll("libmongoose","./lib/")

constant _mg_start_                  = define_c_func(libmongoose_lib, "mg_start", {C_POINTER, C_POINTER, C_POINTER}, C_POINTER)
constant _mg_stop_                   = define_c_proc(libmongoose_lib, "mg_stop", {C_POINTER})
constant _mg_get_option_             = define_c_func(libmongoose_lib, "mg_get_option", {C_POINTER, C_POINTER}, C_POINTER)
constant _mg_get_valid_option_names_ = define_c_func(libmongoose_lib, "mg_get_valid_option_names", {}, C_POINTER)
constant _mg_modify_passwords_file_  = define_c_func(libmongoose_lib, "mg_modify_passwords_file", {C_POINTER, C_POINTER, C_POINTER, C_POINTER}, C_INT)
constant _mg_write_                  = define_c_func(libmongoose_lib, "mg_write", {C_POINTER, C_POINTER, C_POINTER}, C_INT)
constant _mg_send_file_              = define_c_proc(libmongoose_lib, "mg_send_file", {C_POINTER, C_POINTER})
constant _mg_read_                   = define_c_func(libmongoose_lib, "mg_read", {C_POINTER, C_POINTER, C_POINTER}, C_INT)
constant _mg_get_header_             = define_c_func(libmongoose_lib, "mg_get_header", {C_POINTER, C_POINTER}, C_POINTER)
constant _mg_get_var_                = define_c_func(libmongoose_lib, "mg_get_var", {C_POINTER, C_POINTER, C_POINTER, C_POINTER, C_POINTER}, C_INT)
constant _mg_get_cookie_             = define_c_func(libmongoose_lib, "mg_get_cookie", {C_POINTER, C_POINTER, C_POINTER, C_POINTER}, C_INT)
constant _mg_version_                = define_c_func(libmongoose_lib, "mg_version", {}, C_POINTER)

enum 

	false  = 0,
	true

constant

	MESSAGE_FAILED_TO_START = "failed to start mongoose embedded web server",
	MESSAGE_START_LISTEN    = """
----
mongoose embedded web server listening on port(s) %s
press ESC key to terminate server ... """

constant PLUGIN_REQUEST_HANDLER = "request_handler"

sequence route_names = {}, route_handlers = {}, status_code_handlers = {}
integer routes_available = true, code_handlers_available = true

public constant 

	STANDARD_TEXT_REPLY = 

	  "HTTP/1.1 200 OK\r\n"
	& "Content-Type: text/plain\r\n"
	& "Connection: close\r\n\r\n",

	STANDARD_HTML_REPLY =

	  "HTTP/1.1 200 OK\r\n"
	& "Content-Type: text/html\r\n"
	& "Connection: close\r\n\r\n",

	STANDARD_PAGE_NOT_FOUND_REPLY = "404 - Page Not Found",

	STANDARD_ERROR_REPLY = "HTTP/1.1 %d XX\r\nConnection: close\r\n\r\nerror[%d]"

public enum

	IS_LOCAL,
	IS_PLUGIN

public enum

	ROUTE_LIB = 1,
	ROUTE_NAME,
	ROUTE_ID,
	ROUTE_TYPE,
	ROUTE_SHUTDOWN

public enum

	KEY = 1,
	VAL

constant NULL = #0

/* config options */

constant config_options = 
{
  { "C", "cgi_pattern", "**.cgi$|**.pl$|**.php$" },
  { "E", "cgi_environment", NULL },
  { "G", "put_delete_passwords_file", NULL },
  { "I", "cgi_interpreter", NULL },
  { "P", "protect_uri", NULL },
  { "R", "authentication_domain", "mydomain.com" },
  { "S", "ssi_pattern", "**.shtml$|**.shtm$" },
  { "a", "access_log_file", NULL },
  { "c", "ssl_chain_file", NULL },
  { "d", "enable_directory_listing", "yes" },
  { "e", "error_log_file", NULL },
  { "g", "global_passwords_file", NULL },
  { "i", "index_files", "index.html,index.htm,index.cgi" },
  { "k", "enable_keep_alive", "no" },
  { "l", "access_control_list", NULL },
  { "M", "max_request_size", "16384" },
  { "m", "extra_mime_types", NULL },
  { "p", "listening_ports", "8080" },
  { "r", "document_root",  "." },
  { "s", "ssl_certificate", NULL },
  { "t", "num_threads", "10" },
  { "u", "run_as_user", NULL },
  { "w", "url_rewrite_patterns", NULL }
}

/* #define */

/* enum */

--// mg_event
--//
--// Various events on which user-defined function is called by Mongoose.

public enum 

  MG_NEW_REQUEST = 0,   --// New HTTP request has arrived from the client
  MG_HTTP_ERROR,        --// HTTP error must be returned to the client
  MG_EVENT_LOG,         --// Mongoose logs an event, request_info.log_message
  MG_INIT_SSL,          --// Mongoose initializes SSL. Instead of mg_connection *,
                        --// SSL context is passed to the callback function.
  MG_REQUEST_COMPLETE   --// Mongoose has finished handling the request

/* struct */

public memstruct cfg_option
	pointer char short_name
	pointer char name
	pointer char val
end memstruct

--// HTTP headers
public memstruct mg_header 
	pointer char name     --// HTTP header name
	pointer char val      --// HTTP header value
end memstruct

--// This structure contains information about the HTTP request.
public memstruct mg_request_info
	pointer int  user_data          --// User-defined pointer passed to mg_start()
	pointer char request_method     --// "GET", "POST", etc
	pointer char uri                --// URL-decoded URI
	pointer char http_version       --// E.g. "1.0", "1.1"
	pointer char query_string       --// URL part after '?' (not including '?') or NULL
	pointer char remote_user        --// Authenticated user, or NULL if no auth used
	pointer char log_message        --// Mongoose error log message, MG_EVENT_LOG only
	long remote_ip                  --// Client's IP address
	int  remote_port                --// Client's port
	int  status_code                --// HTTP reply status code, e.g. 200
	int  is_ssl                     --// 1 if SSL-ed, 0 if not
	int  num_headers                --// Number of headers
	int  http_headers[64]           --// Maximum 64 headers
	--// ERROR : euc compilation error
	--//
	--// $ euc server/embed.ex
	-- mg_header http_headers[64]   --// http_headers[64], Maximum 64 headers
end memstruct

function ri_string(atom addr)
	if addr > 0 then
		return peek_string(addr)
	end if
	return ""
end function

/* exports */

public function start(atom callback, atom user_data, sequence seq_options)
    object result
    atom options
    options = allocate_string_pointer_array(seq_options)
    result = c_func(_mg_start_, { callback, user_data, options })
	 free_pointer_array(options)
    return result
end function

public procedure stop(atom conn)
    c_proc(_mg_stop_, { conn })
end procedure

public function get_option(atom ctx, sequence seq_name)
    object result
    atom name
    name = allocate_string(seq_name)
    result = c_func(_mg_get_option_, {ctx, name})
    free(name)
	 if result > 0 then
		 return peek_string( result )
	 end if
	 return ""
end function

public function get_valid_option_names()
    return c_func(_mg_get_valid_option_names_, {})
end function

public function modify_passwords_file(sequence seq_passwords_file_name, sequence seq_domain, sequence seq_user, sequence seq_password)
    object result
    atom passwords_file_name, domain, user, password
    passwords_file_name = allocate_string(seq_passwords_file_name)
    domain = allocate_string(seq_domain)
    user = allocate_string(seq_user)
    password = allocate_string(seq_password)
    result = c_func(_mg_modify_passwords_file_, {passwords_file_name, domain, user, password})
    free(passwords_file_name)
    free(domain)
    free(user)
    free(password)
    return result
end function

public function write(atom conn, sequence s)
	integer n = 0
	if conn > 0 then
		atom buf = allocate_string(s)
		n = c_func(_mg_write_, { conn, buf, length(s) })
		free(buf)
	end if
	return n
end function

public procedure send_file(atom conn, sequence seq_path)
    atom path
    path = allocate_string(seq_path)
    c_proc(_mg_send_file_, {conn, path})
	 free(path)
end procedure

public function read(atom conn, atom buf, atom len)
    return c_func(_mg_read_, { conn, buf, len })
end function

public function get_header(atom conn, sequence seq_name)
    object result
    atom name
    name = allocate_string(seq_name)
    result = c_func(_mg_get_header_, { conn, name })
    free(name)
	 if result > 0 then
		 return peek_string( result )
	 end if
	 return ""
end function

public function get_var(sequence seq_data, sequence seq_var_name, atom var_len=1024)
	if not var_len then
		var_len = 1024
	end if
	sequence s = ""
	atom data = allocate_string(seq_data)
	atom var_name = allocate_string(seq_var_name)
	atom buf = allocate(var_len)
	mem_set(buf, NULL, var_len)
	object result = c_func(_mg_get_var_, { data, length(seq_data), var_name, buf, var_len })
	free(data)
	free(var_name)
	s = peek({ buf, var_len })
	free(buf)
	return s
end function

public function get_cookie(atom conn, sequence seq_cookie_name, atom cookie_len=1024)
	if not cookie_len then
		cookie_len = 1024
	end if
	atom cookie_name = allocate_string(seq_cookie_name)
	atom buf = allocate(cookie_len)
	mem_set(buf, NULL, cookie_len)
	object result = c_func(_mg_get_cookie_, { conn, cookie_name, buf, cookie_len})
	free(cookie_name)
	free(buf)
	if result > 0 then
		return peek_string(result)
	end if
	return ""
end function

public function version()
    object result = c_func(_mg_version_, {})
    return peek_string( result )
end function

public function get_options(atom ctx=0)
	sequence opts = {}
	atom addr = get_valid_option_names()
	integer n = sizeof(cfg_option)
	while peek(addr) != #0 do
		sequence k1 = peek_string(peek4u(addr + 0)) --// config short name
		sequence k2 = peek_string(peek4u(addr + 4)) --// config name
		sequence v = ""
		if ctx > 0 then
			v = get_option(ctx, k1) --// config value
		end if
		opts &= {{ k1, v }}
		opts &= {{ k2, v }}
		addr += n
	end while
	return opts
end function

public function wprintf(atom conn, sequence s)
	return write(conn, s)
end function

public function kvget(object kv, object key, object kvdefault="")
/*
	TODO

	expand this routine to access other key value data 
	stores dependent on compilation flags

	eg:

		ifdef __KVSTORE_EUI_MAP__ then
			return map:get(kv, key)
		elsifdef __KVSTORE_JSON__ then
		elsifdef __KVSTORE_ELASTICSEARCH__ then
		elsifdef __KVSTORE_HYPERDEX__ then
		elsifdef __KVSTORE_COUCHDB__ then
		elsifdef __KVSTORE_MONGODB__ then

		etc ...
*/
	for i = 1 to length(kv) do
		if equal(kv[i][KEY], key) then
			return kv[i][VAL]
		end if
	end for
	return kvdefault
end function

public function parse_headers(atom ri)
	sequence headers = {}
	if ri > 0 then
		sequence kv = ri.mg_request_info.http_headers
		for i = 2 to (ri.mg_request_info.num_headers * 2) by 2 do
			headers &=
			{
				{
					peek_string(kv[i - 1]), --// name
					peek_string(kv[i])      --// value
				}
			}
		/*
			--// ERROR : euc compilation error
			--//
			--// $ euc server/embed.ex

			sequence kv = ri.mg_request_info.http_headers[i]
			headers &= 
			{
				{
					peek_string(kv[1]), --// name
					peek_string(kv[2])  --// value
				}
			}
		*/
		end for
	end if
	return headers
end function

public function parse_request(atom ri, integer want_headers=true)
	sequence rqi = {}
	if ri > 0 then
		rqi &= {{ "user_data",      ri.mg_request_info.user_data }}
		rqi &= {{ "request_method", ri_string(ri.mg_request_info.request_method) }}
		rqi &= {{ "uri",            ri_string(ri.mg_request_info.uri) }}
		rqi &= {{ "http_version",   ri_string(ri.mg_request_info.http_version) }}
		rqi &= {{ "query_string",   ri_string(ri.mg_request_info.query_string) }}
		rqi &= {{ "remote_user",    ri_string(ri.mg_request_info.remote_user) }}
		rqi &= {{ "log_message",    ri_string(ri.mg_request_info.log_message) }}
		rqi &= {{ "remote_ip",      ri.mg_request_info.remote_ip }}
		rqi &= {{ "remote_ip_addr", long_to_ipaddr(ri.mg_request_info.remote_ip) }}
		rqi &= {{ "remote_port",    ri.mg_request_info.remote_port }}
		rqi &= {{ "status_code",    ri.mg_request_info.status_code }}
		rqi &= {{ "is_ssl",         ri.mg_request_info.is_ssl }}
		rqi &= {{ "num_headers",    ri.mg_request_info.num_headers }}
		if want_headers then
			rqi &= {{ "http_headers", parse_headers(ri) }}
		end if
	end if
	return rqi
end function

public procedure display_config(atom ctx)
	if ctx > 0 then
		printf(1, "---- Server Configuration\n")
		sequence opts = get_options(ctx)
		for i = 1 to length(opts) do
			if length(opts[i][1]) > 1 then
				printf(1, "%-30s = %s\n", { opts[i][1], opts[i][2] })
			end if
		end for
	end if
end procedure

ifdef __MONGOOSE_WEBSERVER__ then

include std/task.e

constant QUIT_KEY = 27

public procedure load_routes(sequence server_routes)

	--// load shared library plugin route handlers
	route_names          = {}
	route_handlers       = {}
	status_code_handlers = {}

	integer n = length(server_routes)

	if n > 0 then

		for i = 1 to n do

			sequence rh = server_routes[i]
			if (length(rh) != 2) or not sequence(rh[VAL]) then
				continue
			end if

			if integer(rh[KEY]) then
				--// error/status code handlers
				atom shid = routine_id(rh[VAL])
				if shid >= 0 then
					status_code_handlers &= {{ rh[KEY], shid }}
				end if
			else

				--// htm/data route handlers

				atom rhlib     = -1
				atom rhid      = -1
				integer rhtype = IS_LOCAL
				atom rhexit    = -1

				sequence rhandler_name = rh[VAL]

				if not length(rh[KEY]) then
					--// load local route
					rhid = routine_id(rh[VAL]) 
				else
					--// load plugin route
					atom plugin_lib = open_dll(rh[KEY] & "." & SOEXT)
					atom rhidmain = define_c_proc(plugin_lib, rhandler_name, { C_POINTER, C_POINTER })
					if rhidmain >= 0 then
						--// init route handler
						rhid = define_c_proc(plugin_lib, rhandler_name & "_init", {})
						if rhid > 0 then
							c_proc(rhid, {})
						end if
						rhlib  = plugin_lib
						rhid   = rhidmain
						rhtype = IS_PLUGIN
						rhexit = define_c_proc(plugin_lib, rhandler_name & "_exit", {})
					end if
				end if

				if rhid >= 0 then
					--// add route handler callback
					route_handlers &= {{ rhlib, rhandler_name, rhid, rhtype, rhexit }}
					--// update route names
					route_names &= { rh[VAL] }
				end if

			end if

		end for

	end if

end procedure

function route_callback(integer event, atom conn, atom ri)
	switch event do
		case MG_HTTP_ERROR then
			--// error code
			integer status_code = ri.mg_request_info.status_code
			atom rid = kvget(status_code_handlers, status_code, -1)
			if code_handlers_available and rid >= 0 then
				call_proc(rid, { conn, ri })
			else
				write(conn, sprintf(STANDARD_ERROR_REPLY, { status_code, status_code }))
			end if
			return 1
		case MG_NEW_REQUEST then
			--// new request
			if routes_available then
				sequence req_path  = peek_string(ri.mg_request_info.uri)
				integer  route_pos = find(req_path[2..$], route_names)
				if route_pos > 0 then
					switch route_handlers[route_pos][ROUTE_TYPE] do
						case IS_LOCAL then
							--// parse request header information 
							sequence rqi = parse_request(ri, true)
							--// call route handler
							call_proc(route_handlers[route_pos][ROUTE_ID], { conn, rqi })
							break
						case IS_PLUGIN then
							--// call external shared library route handler
							c_proc(route_handlers[route_pos][ROUTE_ID], { conn, ri })
							break
					end switch
					return 1
				end if
			end if
		case MG_REQUEST_COMPLETE then
			break
		case MG_EVENT_LOG then
			break
		case MG_INIT_SSL then
			break
	end switch
	return 0
end function

public procedure listen(sequence start_options, integer show_config=false)

	--// check status code callbacks
	if not length(status_code_handlers) then
		code_handlers_available = false
	end if

	--// check route handler callbaks
	if not length(route_names) or not length(route_handlers) then
		routes_available = false
	end if

	--// init webserver context
	atom callback_id = call_back("+" & routine_id("route_callback"))
	atom ctx         = start(callback_id, NULL, start_options)

	if (ctx = NULL) then
		printf(1, MESSAGE_FAILED_TO_START & "\n")
		abort(0)
	end if

	--// display startup config options
	if show_config then
		display_config(ctx)
		printf(1, MESSAGE_START_LISTEN, { get_option(ctx, "listening_ports") })
	end if

	--// wait for user to terminate server
	while wait_key() != QUIT_KEY do
		task_delay(1)
	end while

	--// clean up
	stop(ctx)

	--// call route handler exit routines
	for i = 1 to length(route_handlers) do
		object lib = route_handlers[i][ROUTE_LIB]
		if atom(lib) then
			atom rhid = route_handlers[i][ROUTE_SHUTDOWN]
			if rhid >= 0 then
				c_proc(rhid, {})
			end if
		end if
	end for

end procedure

end ifdef --// __MONGOOSE_WEBSERVER__

ifdef __TEST_MONGOOSE_IMPORTS__ then
	printf(1, "%-40s = %s\n", { "__MODULENAME__", __MODULENAME__})
	printf(1, "%-40s = %d\n", { "libmongoose_lib", libmongoose_lib})
	printf(1, "%-40s = %d\n", { "_mg_start_", _mg_start_})
	printf(1, "%-40s = %d\n", { "_mg_stop_", _mg_stop_})
	printf(1, "%-40s = %d\n", { "_mg_get_option_", _mg_get_option_})
	printf(1, "%-40s = %d\n", { "_mg_get_valid_option_names_", _mg_get_valid_option_names_})
	printf(1, "%-40s = %d\n", { "_mg_modify_passwords_file_", _mg_modify_passwords_file_})
	printf(1, "%-40s = %d\n", { "_mg_write_", _mg_write_})
	printf(1, "%-40s = %d\n", { "_mg_send_file_", _mg_send_file_})
	printf(1, "%-40s = %d\n", { "_mg_read_", _mg_read_})
	printf(1, "%-40s = %d\n", { "_mg_get_header_", _mg_get_header_})
	printf(1, "%-40s = %d\n", { "_mg_get_var_", _mg_get_var_})
	printf(1, "%-40s = %d\n", { "_mg_get_cookie_", _mg_get_cookie_})
	printf(1, "%-40s = %d\n", { "_mg_version_", _mg_version_})
	printf(1, "%-40s = %s\n", { __MODULENAME__ & " version", version() })
end ifdef
