/*
	source : OpenEuphoria forum post
*/
ifdef WIN32 then
    
    include std/dll.e
    include std/machine.e
    include std/os.e
    
    atom kernel32 = open_dll( "kernel32.dll" )
    
    function lookup_c_func( atom lib, sequence name, sequence params, atom return_type )
        -- search for a series of compatible functions
        
        atom id
        
        if atom( name[1] ) then
            name = {name}
        end if
        
        for i = 1 to length( name ) do
            id = define_c_func( lib, name[i], params, return_type )
            if id != -1 then
                exit
            end if
        end for
        
        return id
    end function
    
    constant
        xGetCurrentProcessId    = define_c_func( kernel32, "GetCurrentProcessId", {}, C_LONG ),
        xGetProcessMemoryInfo   = lookup_c_func( kernel32, {"GetProcessMemoryInfo","K32GetProcessMemoryInfo"}, {C_LONG,C_POINTER,C_LONG}, C_UINT ),
        xOpenProcess            = define_c_func( kernel32, "OpenProcess", {C_LONG,C_UINT,C_LONG}, C_LONG ),
        xCloseHandle            = define_c_proc( kernel32, "CloseHandle", {C_LONG} )
    
    constant
        PROCESS_CREATE_PROCESS              = #0080,
        PROCESS_CREATE_THREAD               = #0002,
        PROCESS_DUP_HANDLE                  = #0040,
        PROCESS_QUERY_INFORMATION           = #0400,
        PROCESS_QUERY_LIMITED_INFORMATION   = #1000,
        PROCESS_SET_INFORMATION             = #0200,
        PROCESS_SET_QUOTA                   = #0100,
        PROCESS_SUSPEND_RESUME              = #0800,
        PROCESS_TERMINATE                   = #0001,
        PROCESS_VM_OPERATION                = #0008,
        PROCESS_VM_READ                     = #0010,
        PROCESS_VM_WRITE                    = #0020,
        PROCESS_ALL_ACCESS                  = #1FCE
    
    constant -- PROCESS_MEMORY_COUNTERS
        memcb                           =  0,
        memPageFaultCount               =  4,
        memPeakWorkingSetSize           =  8,
        memWorkingSetSize               = 12,
        memQuotaPeakPagedPoolUsage      = 16,
        memQuotaPagedPoolUsage          = 20,
        memQuotaPeakNonPagedPoolUsage   = 24,
        memQuotaNonPagedPoolUsage       = 28,
        memPagefileUsage                = 32,
        memPeakPagefileUsage            = 36,
        SIZEOF_PROCESS_MEMORY_COUNTERS  = 40
    
    export function process_memory_usage()
        
        atom processID, hProcess, pmc, result
        integer sizeof_pmc
        
        result = 0
        
        processID = c_func( xGetCurrentProcessId, {} )
        
        hProcess = c_func( xOpenProcess, {or_bits(PROCESS_QUERY_INFORMATION, PROCESS_VM_READ), 0, processID} )
        if hProcess = 0 then
            return -1
        end if
        
        pmc = allocate( SIZEOF_PROCESS_MEMORY_COUNTERS )
        sizeof_pmc = SIZEOF_PROCESS_MEMORY_COUNTERS
        poke4( pmc + memcb, sizeof_pmc )
        
        if c_func( xGetProcessMemoryInfo, {hProcess, pmc, sizeof_pmc} ) then
            result = peek4u( pmc + memPagefileUsage )
        end if
        
        c_proc( xCloseHandle, {hProcess} )
        
        return result
    end function
    
elsifdef LINUX then
    
    include std/get.e
    include std/sequence.e
    
    constant -- /proc/pid/stat fields
        stat_pid        =  1,
        stat_comm       =  2,
        stat_state      =  3,
        stat_ppid       =  4,
        stat_pgrp       =  5,
        stat_session    =  6,
        stat_tty_nr     =  7,
        stat_tpgid      =  8,
        stat_flags      =  9,
        stat_minflt     = 10,
        stat_cminflt    = 11,
        stat_majflt     = 12,
        stat_cmajflt    = 13,
        stat_utime      = 14,
        stat_stime      = 15,
        stat_cutime     = 16,
        stat_cstime     = 17,
        stat_priority   = 18,
        stat_nice       = 19,
        stat_num_threads= 20,
        stat_itrealvalue= 21,
        stat_starttime  = 22,
        stat_vsize      = 23,
        stat_rss        = 24,
        stat_rsslim     = 25,
        stat_startcode  = 26,
        stat_endcode    = 27,
        stat_startstack = 28,
        stat_kstkesp    = 29,
        stat_kstkeip    = 30,
        stat_signal     = 31,
        stat_blocked    = 32,
        stat_sigignore  = 33,
        stat_sigcatch   = 34,
        stat_wchan      = 35,
        stat_nsawp      = 36,
        stat_cnswap     = 37,
        stat_exit_signal= 38,
        stat_processor  = 39,
        stat_rt_priority= 40,
        stat_policy     = 41,
        stat_delayacct_blkio_ticks = 42,
        stat_guest_time = 43,
        stat_cguest_time= 44
    
    export function process_memory_usage()
        
        integer fn
        atom result
        object line
        sequence parts, retval
        
        result = 0
        
        fn = open( "/proc/self/stat", "r" )
        if fn = -1 then
            return -1
        end if
        
        line = gets( fn )
        if sequence( line ) then
            
            parts = split( line, ' ' )
            
            retval = value( parts[stat_vsize] )
            if retval[1] = GET_SUCCESS then
                result = retval[2]
            end if
            
        end if
        
        close( fn )
        
        return result
    end function
    
end ifdef

