/*
	desc      : Mongoose Embedded Web Server Test Application
	author    : Raphael Smart
	date      : May 2012

	licence/disclaimer:

	This software is provided 'as-is', without any express or implied warranty.
	In no event will the author be held liable for any damages arising from
	the use of this software.
*/
--// need this to enable acces to web server startup routines
with define __MONGOOSE_WEBSERVER__

include std/machine.e

include lib/libmongoose.e as mg

enum

	false = 0,
	true   

constant NULL = #0

/* MUST be global */
global procedure test_get_header(atom conn, sequence rqi)
	sequence headers = mg:kvget(rqi, "http_headers")
	printf(1, "HTTP headers : %d\n", { length(headers) })
	for i = 1 to length(headers) do
		printf(1,"[%s] = [%s]\n", { headers[i][KEY], headers[i][VAL] })
	end for
	sequence val = mg:get_header(conn, "Host")
	if length(val) then
		mg:write(conn, sprintf("Value: %s", { val }))
	end if
end procedure

/* MUST be global */
global procedure test_get_var(atom conn, sequence rqi)
	sequence data = ""
	object cl = mg:get_header(conn, "Content-Length")
	sequence query_string = mg:kvget(rqi, "query_string")

	if 
		 find(mg:kvget(rqi, "request_method"), { "POST", "PUT" }) 
	and length(cl) > 0 then

		--// get var from PUT/POST body

		cl  = value(cl)
		cl  = cl[$]
		atom buf = allocate(cl)

      --// read in two pieces, to test continuation
      if (cl > 2) then
         mg:read(conn, buf, 2)
         mg:read(conn, buf + 2, cl - 2)
      else
         mg:read(conn, buf, cl)
      end if
		data = peek({ buf, cl })
		free(buf)

   elsif length(query_string) then
		--// get var from GET
      data = query_string
   end if

   mg:write(conn, mg:STANDARD_HTML_REPLY)
   sequence var = mg:get_var(data, "test_var")
   mg:write(conn, "<h1>test_var=" & var & "</h1> (change test_var, click refresh to test)")

end procedure

/* MUST be global */
global procedure test_get_request_info(atom conn, sequence rqi)
	mg:write(conn, mg:STANDARD_TEXT_REPLY)
	mg:write(conn, sprintf("Method: [%s]\n",       { mg:kvget(rqi, "request_method") }))
	mg:write(conn, sprintf("URI: [%s]\n",          { mg:kvget(rqi, "uri") }))
	mg:write(conn, sprintf("HTTP version: [%s]\n", { mg:kvget(rqi, "http_version") }))
	sequence headers = mg:kvget(rqi, "http_headers")
	for i = 1 to length(headers) do
		mg:write(conn, sprintf("[%s] = [%s]\n", { headers[i][KEY], headers[i][VAL] }))
	end for
	mg:write(conn, sprintf("Query string: [%s]\n",      { mg:kvget(rqi, "query_string") }))
	mg:write(conn, sprintf("Remote IP: [%d]\n",         { mg:kvget(rqi, "remote_ip") }))
	mg:write(conn, sprintf("Remote IP Address: [%s]\n", { mg:kvget(rqi, "remote_ip_addr") }))
	mg:write(conn, sprintf("Remote port: [%d]\n",       { mg:kvget(rqi, "remote_port") }))
	mg:write(conn, sprintf("Remote user: [%s]\n",       { mg:kvget(rqi, "remote_user") }))
	mg:write(conn, "\n(change query string, click refresh to test)")
end procedure

/* MUST be global */
global procedure test_post(atom conn, sequence rqi)
	mg:write(conn, mg:STANDARD_TEXT_REPLY)
	object cl = mg:get_header(conn, "Content-Length")
	if 
		 equal(mg:kvget(rqi, "request_method"), "POST") 
	and length(cl) > 0 then
		cl = value(cl)
		cl = cl[$]
		atom buf = allocate(cl)
		if (buf != NULL) then
			mg:write(conn, buf)
			free(buf)
		end if
	end if
end procedure

constant ERROR_404_PAGE_NOT_FOUND = "www/404.htm"

/* MUST be global */
global procedure status_404_page_not_found(atom conn, atom ri)
	mg:send_file(conn, ERROR_404_PAGE_NOT_FOUND)
end procedure

--// main

	constant server_routes =
	{
		--// local routes
		--// no shared library, route name
		{ "", "test_get_header" },
		{ "", "test_get_var" },
		{ "", "test_get_request_info" },
		{ "", "test_post" },

		--// plugin routes 
		--// shared library, route name
		{ "./server/c_plugin",   "test_c_plugin" },
		{ "./server/mtml_plugin","test_mtml_plugin" },

		--// status code routes
		--// HTTP error status code, error route handler
		{ 404, "status_404_page_not_found" }
	}

	constant start_options = 
	{ 
		"index_files",              "index.htm",
		"listening_ports",          "8080",
		"error_log_file",           "error.log",
		"access_log_file",          "access.log",
		"enable_directory_listing", "yes",
		"authentication_domain",    "",
		"document_root",            "www",
		"cgi_pattern",              "**.cgi$|**.ex$"
	}

	--// load route handlers
	mg:load_routes(server_routes)

	--// start server
	mg:listen(start_options, true)
